#!/usr/bin/perl -w
#       
#       Author: Arlo Randall
#	File:	terminiScanning.pl
#	Date:	11-29-2012
#	Description:
#		This script scans the sequence of a polymorphic antigen and splits it into 
#		N-terminal and C-terminal regions at each possible breakpoint.
#		The sequence variation of each terminal region is then compared to experimental
#		cross-reactivity data for the same set of sequence variants.  
#		The correlation between the global sequence identity matrix and the cross-reactivity
#		maxtrix serves as a baseline correlation value.
#
# 	        For a set of variants of the same antigen two types of data are needed to perform the analysis:
#			1. a multiple sequence alignment
#			2. experimentally determined immune reactivity data from individuals exposed 
#			    to at least one of the variants
#
#	Input Parameters:
#		[0] multiple sequence alignment (1 sequence per line)	# example: msa.txt
#		[1] reactivity file					# example: human_reactivity_n55
#		[2] transform type (enter one of: log rank binary none)
#	Ouput:
#		Results are printed to standard output in tab delimited format with one row
#		for each possible breakpoint. Rows consist of three columns:
#			1. MSA index of break point
#			2. correlation between variation in N-terminal region and cross reactivity
#			3. correlation between variation in C-terminal region and cross reactivity
#		
#----------------------------------------------------------------------------------------

if ($#ARGV != 2) {	
	print "usage:   perl terminiScanning.pl <MSA_file> <reactivity_file> transform_type \n";
        print "example: perl terminiScanning.pl msa.txt human_reactivity_n55.txt log > termini_results.txt \n";
	exit(); 
}

$msa_file = $ARGV[0];
$react_file = $ARGV[1];
$transform_type = lc($ARGV[2]);

# error check for transform type
if($transform_type ne "log" and $transform_type ne "rank" and $transform_type ne "binary" and $transform_type ne "none") {		
	print " Error: $transform_type is not a recognized transform type\n";
        print "   Valid transform types are: log rank binary none\n";
        exit();
}


# read in and process multiple sequence alignment
open(IN,$msa_file);
@text = <IN>;
close(IN);

$max_len = 0;
@names = (); # if MSA is provided in fasta format, this wills save names
@msa_data = ();
foreach $line (@text) {

	$temp = uc($line);
	chomp($temp);
			
	if($temp =~ /\>/) {
		push(@names,$temp);
	}
	elsif($temp =~ /[A-Z]/) {
		@row = split(//,$temp);
		push(@msa_data,[@row]);		
		if(length($temp)>$max_len) {
			$max_len = length($temp);
		}		
	}		
}

# scan MSA and identify variable positions	
@variable_positions = ();
$sequence_count = $#msa_data+1;
for $pos_idx(0..($max_len-1)) {
	# scan specific sequence position	
	$pos_is_variable = 0;
	for $seq_idx(1..($sequence_count-1)) {         
		if($msa_data[$seq_idx][$pos_idx] ne $msa_data[$seq_idx-1][$pos_idx]) {                	
			$pos_is_variable = 1;
		}
	}	
	if($pos_is_variable == 1) {
		push(@variable_positions,$pos_idx);               
	}
}

$var_count = $#variable_positions+1;

# read in and process reactivity data
open(IN,$react_file);
@text = <IN>;
close(IN);

@react_data_raw = ();
foreach $line (@text) {       
	if ($line =~ /[0-9]/) {		
		chomp($line);		
		@row = split(/\s+/,$line);
		push(@react_data_raw,[@row]);				
	}
}

# apply selected transform
@react_data_tf = ();
if($transform_type eq "log") {
	@react_data_tf = &logTransform(\@react_data_raw);
}
elsif($transform_type eq "rank") {
	@react_data_tf = &rankTransform(\@react_data_raw);
}
elsif($transform_type eq "binary") {
	@react_data_tf = &binaryTransform(\@react_data_raw);
}
else{
	@react_data_tf = @react_data_raw; # use raw data
}


@crossReactCorMat = &calcCorrelationMatrix(\@react_data_tf);
@dependent_data = @crossReactCorMat;


@termini_results = ();
for $pos_idx(0..($var_count-2)) {

	# split the sequence into N and C terminal regions
	@n_term_positions = @variable_positions[0 .. $pos_idx];
	@c_term_positions = @variable_positions[$pos_idx+1 .. $var_count-1];

	@identityMatrix = &calcIdentityMatrix(\@msa_data,\@n_term_positions);
	$n_term_r_val = &calcPearsonMatrix(\@identityMatrix,\@crossReactCorMat);

	@identityMatrix = &calcIdentityMatrix(\@msa_data,\@c_term_positions);
	$c_term_r_val = &calcPearsonMatrix(\@identityMatrix,\@crossReactCorMat);

	push(@termini_results,[($variable_positions[$pos_idx],$n_term_r_val,$c_term_r_val)]);
}       

# display results
print "MSA_Pos\tN-Term\tC-Term\n";
for $i(0..$#termini_results) {
	$one_based_idx = $termini_results[$i][0]+1;
	$n_rounded = sprintf("%.3f", $termini_results[$i][1]);
        $c_rounded = sprintf("%.3f", $termini_results[$i][2]);
	print "$one_based_idx\t$n_rounded\t$c_rounded\n";
}

# display global correlation
print "\n";
@identityMatrix = &calcIdentityMatrix(\@msa_data,\@variable_positions);
$global_r_val = &calcPearsonMatrix(\@identityMatrix,\@crossReactCorMat);
print "Global:\n$global_r_val\n";

#**************************************************************************************************
# calculate pearson correlation between two datasets
sub calcPearson()
{      
   	$xref = $_[0];
   	$yref = $_[1];
	@x_vals = @$xref;
	@y_vals = @$yref;
						
        # check that all vals are not exactly the same	
	$x_dif = 0;
	for $i(1..$#x_vals) {
		if($x_vals[$i]!=$x_vals[$i-1]) {
			$x_dif = 1;
		}	
	}
	
	$y_dif = 0;
	for $i(1..$#y_vals) {
		if($y_vals[$i]!=$y_vals[$i-1]) {
			$y_dif = 1;
		}	
	}		
	
	$r_val = 0;
	$N = $#x_vals+1;
	if ($x_dif==1 and $y_dif==1) {
		$x_sum = 0;
		$y_sum = 0;
		$x2_sum = 0;
		$y2_sum = 0;
		$x_sum2 = 0;
		$y_sum2 = 0;
		$xy_sum = 0;	
		for $i(0..$#x_vals) {
			$x = $x_vals[$i];
			$y = $y_vals[$i];
			$x_sum += $x;
			$x2_sum += $x*$x;
			$y_sum += $y;
			$y2_sum += $y*$y;
			$xy_sum += $x*$y;		
		}	
		
		$x_sum2 = $x_sum*$x_sum;
		$y_sum2 = $y_sum*$y_sum;
		
		$numerator = ($N*$xy_sum-$x_sum*$y_sum);
		$denominator = (($N*$x2_sum-$x_sum2)*($N*$y2_sum-$y_sum2));
		if ($denominator != 0) {
			$r_val = $numerator/sqrt($denominator);
		}
	}
	
	return $r_val;

}   
#**************************************************************************************************
# calculate pearson correlation between two datasets
sub calcPearsonMatrix()
{      
   	$xref = $_[0];
   	$yref = $_[1];
	@x_vals_mat = @$xref;
	@y_vals_mat = @$yref;
						
	# if the data are multi dimensional, convert it to a single dimension
	@x_vals = ();
	@y_vals = ();

        # this only works for a matrix with equal numbers of rows and columns		
	for $i(0..$#x_vals_mat) {
		for $j(0..($i-1)) {					
			push(@x_vals,$x_vals_mat[$i][$j]);
			push(@y_vals,$y_vals_mat[$i][$j]);
		}
	}
		

        # check that all vals are not exactly the same	
	$x_dif = 0;
	for $i(1..$#x_vals) {
		if($x_vals[$i]!=$x_vals[$i-1]) {
			$x_dif = 1;
		}	
	}
	
	$y_dif = 0;
	for $i(1..$#y_vals) {
		if($y_vals[$i]!=$y_vals[$i-1]) {
			$y_dif = 1;
		}	
	}		
	
	$r_val = 0;
	$N = $#x_vals+1;
	if ($x_dif==1 and $y_dif==1) {
		$x_sum = 0;
		$y_sum = 0;
		$x2_sum = 0;
		$y2_sum = 0;
		$x_sum2 = 0;
		$y_sum2 = 0;
		$xy_sum = 0;	
		for $i(0..$#x_vals) {
			$x = $x_vals[$i];
			$y = $y_vals[$i];
			$x_sum += $x;
			$x2_sum += $x*$x;
			$y_sum += $y;
			$y2_sum += $y*$y;
			$xy_sum += $x*$y;		
		}	
		
		$x_sum2 = $x_sum*$x_sum;
		$y_sum2 = $y_sum*$y_sum;
		
		$numerator = ($N*$xy_sum-$x_sum*$y_sum);
		$denominator = (($N*$x2_sum-$x_sum2)*($N*$y2_sum-$y_sum2));
		if ($denominator != 0) {
			$r_val = $numerator/sqrt($denominator);
		}
	}
	
	return $r_val;

}   

#**************************************************************************************************
# Apply binary transformation
#    1. Calculate global median
#    2. Set values less or equal to global median to 0
#    3. Set values greater or equal to global median to 1
sub binaryTransform()
{      
   	$data_ref = $_[0];
	@data_raw = @$data_ref;
		
	@data_tf = map { [@$_] } @data_raw;
	
	@all_values = ();
	for $i(0..$#data_tf) {		
		$rowref = $data_tf[$i];
		$cols = @$rowref-1;
		for $j(0..$cols) {								
			push(@all_values,$data_tf[$i][$j]);
		}	
	}

	@sorted_values = sort {$a <=> $b} @all_values;
	$N = $#sorted_values+1;
	
	$median = -1;
	if($N % 2 == 0) {
		$median = ($sorted_values[$N/2]+$sorted_values[$N/2-1])/2.0;
	}
	else {
		$median = $sorted_values[$N/2];
	} 	
	
	for $i(0..$#data_tf) {		
		$rowref = $data_tf[$i];
		$cols = @$rowref-1;
		for $j(0..$cols) {
			if($data_tf[$i][$j] > $median) {
				$data_tf[$i][$j] = 1.0;
			}
			else {
				$data_tf[$i][$j] = 0.0;
			}					
		}	
	}	
	
	return @data_tf;
}  

#**************************************************************************************************
# Apply log10 transformation
sub logTransform()
{      
   	$data_ref = $_[0];
	@data_raw = @$data_ref;		
	@data_tf = map { [@$_] } @data_raw;
	
	for $i(0..$#data_tf) {		
		$rowref = $data_tf[$i];
		$cols = @$rowref-1;
		for $j(0..$cols) {								
			$data_tf[$i][$j] = log($data_tf[$i][$j]);					
		}	
	}
	
	return @data_tf;
}   

#**************************************************************************************************
# Apply rank transformation
# For each antigen variant replace the raw values with the rank in ascending order (i.e. 1 replaces
# the lowest value and N replaces the highest value, where N is the number of samples).

sub rankTransform()
{      
   	$data_ref = $_[0];
	@data_raw = @$data_ref;
		
	@data_tf = map { [@$_] } @data_raw;
	
	for $i(0..$#data_tf) {
		
		$rowref = $data_tf[$i];
		$ncols = @$rowref-1;
		@unsorted_row = ();
		for $j(0..$ncols) {					
			push(@unsorted_row,[($j,$data_tf[$i][$j])]);
		}	
		
		# sort row data on values
		@sorted_row = sort { $a->[1] <=> $b->[1] } @unsorted_row;
		for $j(0..$ncols) {
			$data_tf[$i][$sorted_row[$j][0]] = $j;
		}
		
	}
	
	return @data_tf;
} 

#**************************************************************************************************
# Calculates the identity matrix from a set of sequences using a list of specific positions
sub calcIdentityMatrix()
{
   	$aref = $_[0];
   	$bref = $_[1];
	@msa = @$aref;
	@pos_list = @$bref;
		
	# now compare all pairs of sequences
	@identityMatrix = ();
		
	for $i(0..$#msa) {	
		for $j(0..$#msa) {	
			$matches = 0;
			for $pos_idx(0..$#pos_list) {
				$k = $pos_list[$pos_idx];
				if($msa[$i][$k] eq $msa[$j][$k]) {
					$matches++;		
				}
			}
			$identityMatrix[$i][$j] = $matches;
		}
	}
	
	return @identityMatrix;
}

#**************************************************************************************************
# Calculate Pearson correlation between all pairs of rows and populate correlation matrix
sub calcCorrelationMatrix()
{
   	$aref = $_[0];
	@data = @$aref;
	
	@corMat = ();
		
	foreach $row_x (@data) {
		@row_x_res = ();
		foreach $row_y (@data) {			
			$r_val = &calcPearson(\@$row_x,\@$row_y);
			push(@row_x_res,$r_val);
		}
		push(@corMat,[@row_x_res]);
	}
	
	return @corMat;
}
